@@ -24,7 +24,6 @@
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
-// @conditional=ENABLE(STREAMS_API)
 // @internal
 
 function readableStreamReaderGenericInitialize(reader, stream)
@@ -63,8 +62,26 @@ function privateInitializeReadableStreamDefaultController(stream, underlyingSour
     @putByIdDirectPrivate(this, "pulling", false);
     @putByIdDirectPrivate(this, "strategy", @validateAndNormalizeQueuingStrategy(size, highWaterMark));
 
-    const controller = this;
-    @promiseInvokeOrNoopNoCatch(underlyingSource, "start", [this]).@then(() => {
+    return this;
+}
+
+// https://streams.spec.whatwg.org/#set-up-readable-stream-default-controller, starting from step 6.
+// The other part is implemented in privateInitializeReadableStreamDefaultController.
+function setupReadableStreamDefaultController(stream, underlyingSource, size, highWaterMark, startMethod, pullMethod, cancelMethod)
+{
+    "use strict";
+    const controller = new @ReadableStreamDefaultController(stream, underlyingSource, size, highWaterMark, @isReadableStream);
+    const startAlgorithm = () => @promiseInvokeOrNoopMethodNoCatch(underlyingSource, startMethod, [controller]);
+    const pullAlgorithm = () => @promiseInvokeOrNoopMethod(underlyingSource, pullMethod, [controller]);
+    const cancelAlgorithm = (reason) => @promiseInvokeOrNoopMethod(underlyingSource, cancelMethod, [reason]);
+
+    @putByIdDirectPrivate(controller, "pullAlgorithm", pullAlgorithm);
+    @putByIdDirectPrivate(controller, "cancelAlgorithm", cancelAlgorithm);
+    @putByIdDirectPrivate(controller, "pull", @readableStreamDefaultControllerPull);
+    @putByIdDirectPrivate(controller, "cancel", @readableStreamDefaultControllerCancel);
+    @putByIdDirectPrivate(stream, "readableStreamController", controller);
+
+    startAlgorithm().@then(() => {
         @putByIdDirectPrivate(controller, "started", true);
         @assert(!@getByIdDirectPrivate(controller, "pulling"));
         @assert(!@getByIdDirectPrivate(controller, "pullAgain"));
@@ -72,12 +89,6 @@ function privateInitializeReadableStreamDefaultController(stream, underlyingSour
     }, (error) => {
         @readableStreamDefaultControllerError(controller, error);
     });
-
-    @putByIdDirectPrivate(this, "cancel", @readableStreamDefaultControllerCancel);
-
-    @putByIdDirectPrivate(this, "pull", @readableStreamDefaultControllerPull);
-
-    return this;
 }
 
 function readableStreamDefaultControllerError(controller, error)
@@ -120,6 +131,261 @@ function readableStreamPipeTo(stream, sink)
     doPipe();
 }
 
+function acquireReadableStreamDefaultReader(stream)
+{
+    return new @ReadableStreamDefaultReader(stream);
+}
+
+// FIXME: Replace readableStreamPipeTo by below function.
+// This method implements the latest https://streams.spec.whatwg.org/#readable-stream-pipe-to.
+function readableStreamPipeToWritableStream(source, destination, preventClose, preventAbort, preventCancel, signal)
+{
+    @assert(@isReadableStream(source));
+    @assert(@isWritableStream(destination));
+    @assert(!@isReadableStreamLocked(source));
+    @assert(!@isWritableStreamLocked(destination));
+    @assert(signal === @undefined || signal instanceof @AbortSignal);
+
+    if (@getByIdDirectPrivate(source, "underlyingByteSource") !== @undefined)
+        return @Promise.@reject("Piping of readable by strean is not supported");
+
+    let pipeState = { source : source, destination : destination, preventAbort : preventAbort, preventCancel : preventCancel, preventClose : preventClose, signal : signal };
+
+    pipeState.reader = @acquireReadableStreamDefaultReader(source);
+    pipeState.writer = @acquireWritableStreamDefaultWriter(destination);
+
+    @putByIdDirectPrivate(source, "disturbed", true);
+
+    pipeState.finalized = false;
+    pipeState.shuttingDown = false;
+    pipeState.promiseCapability = @newPromiseCapability(@Promise);
+    pipeState.pendingReadPromiseCapability = @newPromiseCapability(@Promise);
+    pipeState.pendingReadPromiseCapability.@resolve.@call();
+    pipeState.pendingWritePromise = @Promise.@resolve();
+
+    if (signal !== @undefined) {
+        const algorithm = () => {
+            if (pipeState.finalized)
+                return;
+
+            const error = @makeDOMException("AbortError", "abort pipeTo from signal");
+
+            @pipeToShutdownWithAction(pipeState, () => {
+                const shouldAbortDestination = !pipeState.preventAbort && @getByIdDirectPrivate(pipeState.destination, "state") === "writable";
+                const promiseDestination = shouldAbortDestination ? @writableStreamAbort(pipeState.destination, error) : @Promise.@resolve();
+
+                const shouldAbortSource = !pipeState.preventCancel && @getByIdDirectPrivate(pipeState.source, "state") === @streamReadable;
+                const promiseSource = shouldAbortSource ? @readableStreamCancel(pipeState.source, error) : @Promise.@resolve();
+
+                let promiseCapability = @newPromiseCapability(@Promise);
+                let shouldWait = true;
+                let handleResolvedPromise = () => {
+                    if (shouldWait) {
+                        shouldWait = false;
+                        return;
+                    }
+                    promiseCapability.@resolve.@call();
+                }
+                let handleRejectedPromise = (e) => {
+                    promiseCapability.@reject.@call(@undefined, e);
+                }
+                promiseDestination.@then(handleResolvedPromise, handleRejectedPromise);
+                promiseSource.@then(handleResolvedPromise, handleRejectedPromise);
+                return promiseCapability.@promise;
+            }, error);
+        };
+        if (@whenSignalAborted(signal, algorithm))
+            return pipeState.promiseCapability.@promise;
+    }
+
+    @pipeToErrorsMustBePropagatedForward(pipeState);
+    @pipeToErrorsMustBePropagatedBackward(pipeState);
+    @pipeToClosingMustBePropagatedForward(pipeState);
+    @pipeToClosingMustBePropagatedBackward(pipeState);
+
+    @pipeToLoop(pipeState);
+
+    return pipeState.promiseCapability.@promise;
+}
+
+function pipeToLoop(pipeState)
+{
+    if (pipeState.shuttingDown)
+        return;
+
+    @pipeToDoReadWrite(pipeState).@then((result) => {
+        if (result)
+            @pipeToLoop(pipeState);
+    });
+}
+
+function pipeToDoReadWrite(pipeState)
+{
+    @assert(!pipeState.shuttingDown);
+
+    pipeState.pendingReadPromiseCapability = @newPromiseCapability(@Promise);
+    @getByIdDirectPrivate(pipeState.writer, "readyPromise").@promise.@then(() => {
+        if (pipeState.shuttingDown) {
+            pipeState.pendingReadPromiseCapability.@resolve.@call(@undefined, false);
+            return;
+        }
+
+        @readableStreamDefaultReaderRead(pipeState.reader).@then((result) => {
+            const canWrite = !result.done && @getByIdDirectPrivate(pipeState.writer, "stream") !== @undefined;
+            pipeState.pendingReadPromiseCapability.@resolve.@call(@undefined, canWrite);
+            if (!canWrite)
+                return;
+
+            pipeState.pendingWritePromise = @writableStreamDefaultWriterWrite(pipeState.writer, result.value);
+        }, (e) => {
+            pipeState.pendingReadPromiseCapability.@resolve.@call(@undefined, false);
+        });
+    }, (e) => {
+        pipeState.pendingReadPromiseCapability.@resolve.@call(@undefined, false);
+    });
+    return pipeState.pendingReadPromiseCapability.@promise;
+}
+
+function pipeToErrorsMustBePropagatedForward(pipeState)
+{
+    const action = () => {
+        pipeState.pendingReadPromiseCapability.@resolve.@call(@undefined, false);
+        const error = @getByIdDirectPrivate(pipeState.source, "storedError");
+        if (!pipeState.preventAbort) {
+            @pipeToShutdownWithAction(pipeState, () => @writableStreamAbort(pipeState.destination, error), error);
+            return;
+        }
+        @pipeToShutdown(pipeState, error);
+    };
+
+    if (@getByIdDirectPrivate(pipeState.source, "state") === @streamErrored) {
+        action();
+        return;
+    }
+
+    @getByIdDirectPrivate(pipeState.reader, "closedPromiseCapability").@promise.@then(@undefined, action);
+}
+
+function pipeToErrorsMustBePropagatedBackward(pipeState)
+{
+    const action = () => {
+        const error = @getByIdDirectPrivate(pipeState.destination, "storedError");
+        if (!pipeState.preventCancel) {
+            @pipeToShutdownWithAction(pipeState, () => @readableStreamCancel(pipeState.source, error), error);
+            return;
+        }
+        @pipeToShutdown(pipeState, error);
+    };
+    if (@getByIdDirectPrivate(pipeState.destination, "state") === "errored") {
+        action();
+        return;
+    }
+    @getByIdDirectPrivate(pipeState.writer, "closedPromise").@promise.@then(@undefined, action);
+}
+
+function pipeToClosingMustBePropagatedForward(pipeState)
+{
+    const action = () => {
+        pipeState.pendingReadPromiseCapability.@resolve.@call(@undefined, false);
+        const error = @getByIdDirectPrivate(pipeState.source, "storedError");
+        if (!pipeState.preventClose) {
+            @pipeToShutdownWithAction(pipeState, () => @writableStreamDefaultWriterCloseWithErrorPropagation(pipeState.writer));
+            return;
+        }
+        @pipeToShutdown(pipeState);
+    };
+    if (@getByIdDirectPrivate(pipeState.source, "state") === @streamClosed) {
+        action();
+        return;
+    }
+    @getByIdDirectPrivate(pipeState.reader, "closedPromiseCapability").@promise.@then(action, @undefined);
+}
+
+function pipeToClosingMustBePropagatedBackward(pipeState)
+{
+    if (!@writableStreamCloseQueuedOrInFlight(pipeState.destination) && @getByIdDirectPrivate(pipeState.destination, "state") !== "closed")
+        return;
+
+    // @assert no chunks have been read/written
+
+    const error = @makeTypeError("closing is propagated backward");
+    if (!pipeState.preventCancel) {
+        @pipeToShutdownWithAction(pipeState, () => @readableStreamCancel(pipeState.source, error), error);
+        return;
+    }
+    @pipeToShutdown(pipeState, error);
+}
+
+function pipeToShutdownWithAction(pipeState, action)
+{
+    if (pipeState.shuttingDown)
+        return;
+
+    pipeState.shuttingDown = true;
+
+    const hasError = arguments.length > 2;
+    const error = arguments[2];
+    const finalize = () => {
+        const promise = action();
+        promise.@then(() => {
+            if (hasError)
+                @pipeToFinalize(pipeState, error);
+            else
+                @pipeToFinalize(pipeState);
+        }, (e)  => {
+            @pipeToFinalize(pipeState, e);
+        });
+    };
+
+    if (@getByIdDirectPrivate(pipeState.destination, "state") === "writable" && !@writableStreamCloseQueuedOrInFlight(pipeState.destination)) {
+        pipeState.pendingReadPromiseCapability.@promise.@then(() => {
+            pipeState.pendingWritePromise.@then(finalize, finalize);
+        }, (e) => @pipeToFinalize(pipeState, e));
+        return;
+    }
+
+    finalize();
+}
+
+function pipeToShutdown(pipeState)
+{
+    if (pipeState.shuttingDown)
+        return;
+
+    pipeState.shuttingDown = true;
+
+    const hasError = arguments.length > 1;
+    const error = arguments[1];
+    const finalize = () => {
+        if (hasError)
+            @pipeToFinalize(pipeState, error);
+        else
+            @pipeToFinalize(pipeState);
+    };
+
+    if (@getByIdDirectPrivate(pipeState.destination, "state") === "writable" && !@writableStreamCloseQueuedOrInFlight(pipeState.destination)) {
+        pipeState.pendingReadPromiseCapability.@promise.@then(() => {
+            pipeState.pendingWritePromise.@then(finalize, finalize);
+        }, (e) => @pipeToFinalize(pipeState, e));
+        return;
+    }
+    finalize();
+}
+
+function pipeToFinalize(pipeState)
+{
+    @writableStreamDefaultWriterRelease(pipeState.writer);
+    @readableStreamReaderGenericRelease(pipeState.reader);
+
+    // Instead of removing the abort algorithm as per spec, we make it a no-op which is equivalent.
+    pipeState.finalized = true;
+
+    if (arguments.length > 1)
+        pipeState.promiseCapability.@reject.@call(@undefined, arguments[1]);
+    else
+        pipeState.promiseCapability.@resolve.@call();
+}
+
 function readableStreamTee(stream, shouldClone)
 {
     "use strict";
@@ -137,18 +403,20 @@ function readableStreamTee(stream, shouldClone)
         reason2: @undefined,
     };
 
-    teeState.cancelPromiseCapability = @newPromiseCapability(@InternalPromise);
+    teeState.cancelPromiseCapability = @newPromiseCapability(@Promise);
 
     const pullFunction = @readableStreamTeePullFunction(teeState, reader, shouldClone);
 
-    const branch1 = new @ReadableStream({
-        "pull": pullFunction,
-        "cancel": @readableStreamTeeBranch1CancelFunction(teeState, stream)
-    });
-    const branch2 = new @ReadableStream({
-        "pull": pullFunction,
-        "cancel": @readableStreamTeeBranch2CancelFunction(teeState, stream)
-    });
+    const branch1Source = { };
+    @putByIdDirectPrivate(branch1Source, "pull", pullFunction);
+    @putByIdDirectPrivate(branch1Source, "cancel", @readableStreamTeeBranch1CancelFunction(teeState, stream));
+
+    const branch2Source = { };
+    @putByIdDirectPrivate(branch2Source, "pull", pullFunction);
+    @putByIdDirectPrivate(branch2Source, "cancel", @readableStreamTeeBranch2CancelFunction(teeState, stream));
+
+    const branch1 = new @ReadableStream(branch1Source);
+    const branch2 = new @ReadableStream(branch2Source);
 
     @getByIdDirectPrivate(reader, "closedPromiseCapability").@promise.@then(@undefined, function(e) {
         if (teeState.closedOrErrored)
@@ -156,6 +424,7 @@ function readableStreamTee(stream, shouldClone)
         @readableStreamDefaultControllerError(branch1.@readableStreamController, e);
         @readableStreamDefaultControllerError(branch2.@readableStreamController, e);
         teeState.closedOrErrored = true;
+        teeState.cancelPromiseCapability.@resolve.@call();
     });
 
     // Additional fields compared to the spec, as they are needed within pull/cancel functions.
@@ -195,6 +464,7 @@ function readableStreamTeePullFunction(teeState, reader, shouldClone)
                 if (!teeState.canceled2)
                     @readableStreamDefaultControllerClose(teeState.branch2.@readableStreamController);
                 teeState.closedOrErrored = true;
+                teeState.cancelPromiseCapability.@resolve.@call();
             }
             if (teeState.closedOrErrored)
                 return;
@@ -285,26 +555,42 @@ function readableStreamError(stream, error)
 
     if (@isReadableStreamDefaultReader(reader)) {
         const requests = @getByIdDirectPrivate(reader, "readRequests");
-        for (let index = 0, length = requests.length; index < length; ++index)
-            requests[index].@reject.@call(@undefined, error);
         @putByIdDirectPrivate(reader, "readRequests", []);
+        for (let index = 0, length = requests.length; index < length; ++index)
+            @rejectPromise(requests[index], error);
     } else {
         @assert(@isReadableStreamBYOBReader(reader));
         const requests = @getByIdDirectPrivate(reader, "readIntoRequests");
-        for (let index = 0, length = requests.length; index < length; ++index)
-            requests[index].@reject.@call(@undefined, error);
         @putByIdDirectPrivate(reader, "readIntoRequests", []);
+        for (let index = 0, length = requests.length; index < length; ++index)
+            @rejectPromise(requests[index], error);
     }
 
     @getByIdDirectPrivate(reader, "closedPromiseCapability").@reject.@call(@undefined, error);
     const promise = @getByIdDirectPrivate(reader, "closedPromiseCapability").@promise;
-    @putPromiseInternalField(promise, @promiseFieldFlags, @getPromiseInternalField(promise, @promiseFieldFlags) | @promiseFlagsIsHandled);
+    @markPromiseAsHandled(promise);
+}
+
+function readableStreamDefaultControllerShouldCallPull(controller)
+{
+    const stream = @getByIdDirectPrivate(controller, "controlledReadableStream");
+
+    if (!@readableStreamDefaultControllerCanCloseOrEnqueue(controller))
+        return false;
+    if (!@getByIdDirectPrivate(controller, "started"))
+        return false;
+    if ((!@isReadableStreamLocked(stream) || !@getByIdDirectPrivate(@getByIdDirectPrivate(stream, "reader"), "readRequests").length) && @readableStreamDefaultControllerGetDesiredSize(controller) <= 0)
+        return false;
+    const desiredSize = @readableStreamDefaultControllerGetDesiredSize(controller);
+    @assert(desiredSize !== null);
+    return desiredSize > 0;
 }
 
 function readableStreamDefaultControllerCallPullIfNeeded(controller)
 {
     "use strict";
 
+    // FIXME: use @readableStreamDefaultControllerShouldCallPull
     const stream = @getByIdDirectPrivate(controller, "controlledReadableStream");
 
     if (!@readableStreamDefaultControllerCanCloseOrEnqueue(controller))
@@ -322,7 +608,7 @@ function readableStreamDefaultControllerCallPullIfNeeded(controller)
     @assert(!@getByIdDirectPrivate(controller, "pullAgain"));
     @putByIdDirectPrivate(controller, "pulling", true);
 
-    @promiseInvokeOrNoop(@getByIdDirectPrivate(controller, "underlyingSource"), "pull", [controller]).@then(function() {
+    @getByIdDirectPrivate(controller, "pullAlgorithm").@call(@undefined).@then(function() {
         @putByIdDirectPrivate(controller, "pulling", false);
         if (@getByIdDirectPrivate(controller, "pullAgain")) {
             @putByIdDirectPrivate(controller, "pullAgain", false);
@@ -385,7 +671,7 @@ function readableStreamDefaultControllerCancel(controller, reason)
     "use strict";
 
     @putByIdDirectPrivate(controller, "queue", @newQueue());
-    return @promiseInvokeOrNoop(@getByIdDirectPrivate(controller, "underlyingSource"), "cancel", [reason]);
+    return @getByIdDirectPrivate(controller, "cancelAlgorithm").@call(@undefined, reason);
 }
 
 function readableStreamDefaultControllerPull(controller)
@@ -399,7 +685,8 @@ function readableStreamDefaultControllerPull(controller)
             @readableStreamClose(stream);
         else
             @readableStreamDefaultControllerCallPullIfNeeded(controller);
-        return @Promise.@resolve({value: chunk, done: false});
+
+        return @createFulfilledPromise({ value: chunk, done: false });
     }
     const pendingPromise = @readableStreamAddReadRequest(stream);
     @readableStreamDefaultControllerCallPullIfNeeded(controller);
@@ -429,9 +716,9 @@ function readableStreamClose(stream)
 
     if (@isReadableStreamDefaultReader(reader)) {
         const requests = @getByIdDirectPrivate(reader, "readRequests");
-        for (let index = 0, length = requests.length; index < length; ++index)
-            requests[index].@resolve.@call(@undefined, {value:@undefined, done: true});
         @putByIdDirectPrivate(reader, "readRequests", []);
+        for (let index = 0, length = requests.length; index < length; ++index)
+            @fulfillPromise(requests[index], { value: @undefined, done: true });
     }
 
     @getByIdDirectPrivate(reader, "closedPromiseCapability").@resolve.@call();
@@ -440,8 +727,8 @@ function readableStreamClose(stream)
 function readableStreamFulfillReadRequest(stream, chunk, done)
 {
     "use strict";
-
-    @getByIdDirectPrivate(@getByIdDirectPrivate(stream, "reader"), "readRequests").@shift().@resolve.@call(@undefined, {value: chunk, done: done});
+    const readRequest = @getByIdDirectPrivate(@getByIdDirectPrivate(stream, "reader"), "readRequests").@shift();
+    @fulfillPromise(readRequest, { value: chunk, done: done });
 }
 
 function readableStreamDefaultControllerEnqueue(controller, chunk)
@@ -480,7 +767,7 @@ function readableStreamDefaultReaderRead(reader)
 
     @putByIdDirectPrivate(stream, "disturbed", true);
     if (state === @streamClosed)
-        return @Promise.@resolve({value: @undefined, done: true});
+        return @createFulfilledPromise({ value: @undefined, done: true });
     if (state === @streamErrored)
         return @Promise.@reject(@getByIdDirectPrivate(stream, "storedError"));
     @assert(state === @streamReadable);
@@ -495,10 +782,10 @@ function readableStreamAddReadRequest(stream)
     @assert(@isReadableStreamDefaultReader(@getByIdDirectPrivate(stream, "reader")));
     @assert(@getByIdDirectPrivate(stream, "state") == @streamReadable);
 
-    const readRequest = @newPromiseCapability(@Promise);
-    @getByIdDirectPrivate(@getByIdDirectPrivate(stream, "reader"), "readRequests").@push(readRequest);
+    const readRequest = @newPromise();
+    @arrayPush(@getByIdDirectPrivate(@getByIdDirectPrivate(stream, "reader"), "readRequests"), readRequest);
 
-    return readRequest.@promise;
+    return readRequest;
 }
 
 function isReadableStreamDisturbed(stream)
@@ -522,7 +809,7 @@ function readableStreamReaderGenericRelease(reader)
         @putByIdDirectPrivate(reader, "closedPromiseCapability", { @promise: @newHandledRejectedPromise(@makeTypeError("reader released lock")) });
 
     const promise = @getByIdDirectPrivate(reader, "closedPromiseCapability").@promise;
-    @putPromiseInternalField(promise, @promiseFieldFlags, @getPromiseInternalField(promise, @promiseFieldFlags) | @promiseFlagsIsHandled);
+    @markPromiseAsHandled(promise);
     @putByIdDirectPrivate(@getByIdDirectPrivate(reader, "ownerReadableStream"), "reader", @undefined);
     @putByIdDirectPrivate(reader, "ownerReadableStream", @undefined);
 }
